from typing import Optional

import attr

from autoextract_poet.items import Article, Product, ProductList, ArticleList, \
    Comments, ForumPosts, JobPosting, RealEstate, Reviews, Vehicle
from autoextract_poet.page_inputs import AutoExtractHtml, \
    AutoExtractProductData, AutoExtractProductListData, AutoExtractArticleData, \
    T, AutoExtractArticleListData, AutoExtractCommentsData, \
    AutoExtractForumPostsData, AutoExtractJobPostingData, \
    AutoExtractRealEstateData, AutoExtractReviewsData, AutoExtractVehicleData
from autoextract_poet.util import export
from web_poet import Injectable, ItemPage
from web_poet.mixins import ResponseShortcutsMixin


@export
@attr.s(auto_attribs=True)
class AutoExtractWebPage(Injectable, ResponseShortcutsMixin):
    """Base Page Object which requires :class:`~.AutoExtractHtml`
    and provides XPath / CSS shortcuts.

    Use this class as a base class for Page Objects which work on
    the browser HTML provided by AutoExtract.
    """
    response: AutoExtractHtml


@export
@attr.s(auto_attribs=True)
class AutoExtractItemWebPage(AutoExtractWebPage, ItemPage):
    """:class:`AutoExtractWebPage` that requires the :meth:`to_item` method to
    be implemented.
    """
    pass


@export
@attr.s(auto_attribs=True)
class AutoExtractArticlePage(ItemPage):
    """
    Article data from AutoExtract

    https://docs.zyte.com/automatic-extraction/article.html
    """
    article_data: AutoExtractArticleData

    def to_item(self) -> Optional[Article]:
        return self.article_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractArticleListPage(ItemPage):
    """
    Article list data from AutoExtract

    https://docs.zyte.com/automatic-extraction/article-list.html
    """
    article_list_data: AutoExtractArticleListData

    def to_item(self) -> Optional[ArticleList]:
        return self.article_list_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractProductPage(ItemPage):
    """
    Product data from AutoExtract

    https://docs.zyte.com/automatic-extraction/product.html
    """
    product_data: AutoExtractProductData

    def to_item(self) -> Optional[Product]:
        return self.product_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractProductListPage(ItemPage):
    """
    Product list data from AutoExtract

    https://docs.zyte.com/automatic-extraction/product-list.html
    """
    product_list_data: AutoExtractProductListData

    def to_item(self) -> Optional[ProductList]:
        return self.product_list_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractCommentsPage(ItemPage):
    """
    Comments data from AutoExtract

    https://docs.zyte.com/automatic-extraction/comment.html
    """
    comments_data: AutoExtractCommentsData

    def to_item(self) -> Optional[Comments]:
        return self.comments_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractForumPostsPage(ItemPage):
    """
    Forum posts data from AutoExtract

    https://docs.zyte.com/automatic-extraction/forum-post.html
    """
    forum_posts_data: AutoExtractForumPostsData

    def to_item(self) -> Optional[ForumPosts]:
        return self.forum_posts_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractJobPostingPage(ItemPage):
    """
    Job posting data from AutoExtract

    https://docs.zyte.com/automatic-extraction/job-posting.html
    """
    job_posting_data: AutoExtractJobPostingData

    def to_item(self) -> Optional[JobPosting]:
        return self.job_posting_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractRealEstatePage(ItemPage):
    """
    Real estate data from AutoExtract

    https://docs.zyte.com/automatic-extraction/real-estate.html
    """
    real_estate_data: AutoExtractRealEstateData

    def to_item(self) -> Optional[RealEstate]:
        return self.real_estate_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractReviewsPage(ItemPage):
    """
    Reviews data from AutoExtract

    https://docs.zyte.com/automatic-extraction/review.html
    """
    reviews_data: AutoExtractReviewsData

    def to_item(self) -> Optional[Reviews]:
        return self.reviews_data.to_item()


@export
@attr.s(auto_attribs=True)
class AutoExtractVehiclePage(ItemPage):
    """
    Vehicle data from AutoExtract

    https://docs.zyte.com/automatic-extraction/vehicle.html
    """
    vehicle_data: AutoExtractVehicleData

    def to_item(self) -> Optional[Vehicle]:
        return self.vehicle_data.to_item()



from typing import Dict, List, Optional

import attr

from .util import split_in_unknown_and_known_fields


class _ItemBase:
    # Reserving an slot for _unknown_fields_dict.
    # This is done in a base class because otherwise attr.s won't pick it up
    __slots__ = ("_unknown_fields_dict", )


@attr.s(auto_attribs=True, slots=True)
class Item(_ItemBase):

    def __attrs_post_init__(self):
        self._unknown_fields_dict = {}

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        """
        Read an item from a dictionary.

        Unknown attributes are kept in the dict ``_unknown_fields_dict``
        so that ``AutoExtractAdapter`` can include them in the resultant item.
        This ensures supporting new AutoExtract fields even if the
        item library is not in sync.
        """
        if not item:
            return None
        unknown_fields, known_fields = split_in_unknown_and_known_fields(item, cls)
        obj = cls(**known_fields)  # type: ignore
        obj._unknown_fields_dict = unknown_fields
        return obj

    @classmethod
    def from_list(cls, items: Optional[List[Dict]]) -> List:
        """
        Read items from a list, invoking ``from_dict`` for each item in the list
        """
        return [cls.from_dict(item) for item in items or []]


def _apply_types(item: Optional[dict], *,
                 from_dict: dict = None,
                 from_list: dict = None) -> Optional[dict]:
    """
    Utility function to wrap basic types into data structures.

    Assumptions:
    * field names stay the same;
    * from_dict fields are optional, and in case of an absent field
      we're preserving None instead of creating an empty data structure;
    * from_list fields are optional, but in case of an absent field
      an empty list is created.
    """
    if not item:
        return None

    new_item = dict(**item)
    new_item.update({
        key: cls.from_dict(item.get(key))
        for key, cls in (from_dict or {}).items()
    })
    new_item.update({
        key: cls.from_list(item.get(key, []))
        for key, cls in (from_list or {}).items()
    })
    return new_item


@attr.s(auto_attribs=True, slots=True)
class Offer(Item):

    price: Optional[str] = None
    currency: Optional[str] = None
    availability: Optional[str] = None
    regularPrice: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Breadcrumb(Item):

    name: Optional[str] = None
    link: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Rating(Item):

    ratingValue: Optional[float] = None
    bestRating: Optional[float] = None
    reviewCount: Optional[int] = None


@attr.s(auto_attribs=True, slots=True)
class AdditionalProperty(Item):

    name: str
    value: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class GTIN(Item):

    type: str
    value: str


@attr.s(auto_attribs=True, slots=True)
class PaginationLink(Item):

    url: Optional[str] = None
    text: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Article(Item):

    url: Optional[str] = None
    probability: Optional[float] = None
    headline: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    dateModified: Optional[str] = None
    dateModifiedRaw: Optional[str] = None
    author: Optional[str] = None
    authorsList: List[str] = attr.Factory(list)
    inLanguage: Optional[str] = None
    breadcrumbs: List[Breadcrumb] = attr.Factory(list)
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    description: Optional[str] = None
    articleBody: Optional[str] = None
    articleBodyHtml: Optional[str] = None
    articleBodyRaw: Optional[str] = None
    videoUrls: List[str] = attr.Factory(list)
    audioUrls: List[str] = attr.Factory(list)
    canonicalUrl: Optional[str] = None


    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(item, from_list={'breadcrumbs': Breadcrumb})
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class ArticleFromList(Item):

    url: Optional[str] = None
    probability: Optional[float] = None
    headline: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    author: Optional[str] = None
    authorsList: List[str] = attr.Factory(list)
    inLanguage: Optional[str] = None
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    articleBody: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class ArticleList(Item):

    url: Optional[str] = None
    articles: List[ArticleFromList] = attr.Factory(list)
    paginationNext: Optional[PaginationLink] = None
    paginationPrevious: Optional[PaginationLink] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_list={'articles': ArticleFromList},
            from_dict={
                'paginationNext': PaginationLink,
                'paginationPrevious': PaginationLink
            }
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Product(Item):

    url: Optional[str] = None
    canonicalUrl: Optional[str] = None
    probability: Optional[float] = None
    name: Optional[str] = None
    offers: List[Offer] = attr.Factory(list)
    sku: Optional[str] = None
    gtin: List[GTIN] = attr.Factory(list)
    mpn: Optional[str] = None
    brand: Optional[str] = None
    breadcrumbs: List[Breadcrumb] = attr.Factory(list)
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    additionalProperty: List[AdditionalProperty] = attr.Factory(list)
    aggregateRating: Optional[Rating] = None
    color: Optional[str] = None
    size: Optional[str] = None
    style: Optional[str] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_dict={'aggregateRating': Rating},
            from_list={
                'additionalProperty': AdditionalProperty,
                'breadcrumbs': Breadcrumb,
                'gtin': GTIN,
                'offers': Offer,
            }
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class ProductFromList(Item):

    probability: Optional[float] = None
    url: Optional[str] = None
    name: Optional[str] = None
    offers: List[Offer] = attr.Factory(list)
    sku: Optional[str] = None
    brand: Optional[str] = None
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    description: Optional[str] = None
    aggregateRating: Optional[Rating] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_dict={'aggregateRating': Rating},
            from_list={'offers': Offer}
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class ProductList(Item):

    url: Optional[str] = None
    products: List[ProductFromList] = attr.Factory(list)
    breadcrumbs: List[Breadcrumb] = attr.Factory(list)
    paginationNext: Optional[PaginationLink] = None
    paginationPrevious: Optional[PaginationLink] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_list={
                'products': ProductFromList,
                'breadcrumbs': Breadcrumb,
            },
            from_dict={
                'paginationNext': PaginationLink,
                'paginationPrevious': PaginationLink,
            }
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Location(Item):

    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Organization(Item):

    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Salary(Item):
    raw: Optional[str] = None
    value: Optional[float] = None
    currency: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class JobPosting(Item):

    probability: Optional[float] = None
    url: Optional[str] = None
    title: Optional[str] = None
    datePosted: Optional[str] = None
    validThrough: Optional[str] = None
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    employmentType: Optional[str] = None
    hiringOrganization: Optional[Organization] = None
    baseSalary: Optional[Salary] = None
    jobLocation: Optional[Location] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(item, from_dict={
            "hiringOrganization": Organization,
            "baseSalary": Salary,
            "jobLocation": Location,
        })
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Comment(Item):

    probability: Optional[float] = None
    text: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    upvoteCount: Optional[int] = None
    downvoteCount: Optional[int] = None


@attr.s(auto_attribs=True, slots=True)
class Comments(Item):

    url: Optional[str] = None
    comments: List[Comment] = attr.Factory(list)

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(item, from_list={"comments": Comment})
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class ForumPost(Item):

    probability: Optional[float] = None
    text: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    upvoteCount: Optional[int] = None
    downvoteCount: Optional[int] = None
    replyCount: Optional[int] = None


@attr.s(auto_attribs=True, slots=True)
class Topic(Item):

    name: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class ForumPosts(Item):

    url: Optional[str] = None
    topic: Optional[Topic] = None
    posts: List[ForumPost] = attr.Factory(list)

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(item,
                                from_list={"posts": ForumPost},
                                from_dict={"topic": Topic})
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Address(Item):

    postalCode: Optional[str] = None
    streetAddress: Optional[str] = None
    addressCountry: Optional[str] = None
    addressLocality: Optional[str] = None
    addressRegion: Optional[str] = None
    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Area(Item):

    value: Optional[float] = None
    unitCode: Optional[str] = None
    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class TradeAction(Item):

    tradeType: Optional[str] = None
    price: Optional[str] = None
    currency: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class RealEstate(Item):

    url: Optional[str] = None
    probability: Optional[float] = None
    name: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    description: Optional[str] = None
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    yearBuilt: Optional[int] = None
    breadcrumbs: List[Breadcrumb] = attr.Factory(list)
    additionalProperty: List[AdditionalProperty] = attr.Factory(list)
    address: Optional[Address] = None
    area: Optional[Area] = None
    numberOfBathroomsTotal: Optional[int] = None
    numberOfFullBathrooms: Optional[int] = None
    numberOfPartialBathrooms: Optional[int] = None
    numberOfBedrooms: Optional[int] = None
    numberOfRooms: Optional[int] = None
    identifier: Optional[str] = None
    tradeActions: List[TradeAction] = attr.Factory(list)

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_list={
                "additionalProperty": AdditionalProperty,
                "breadcrumbs": Breadcrumb,
                "tradeActions": TradeAction
            },
            from_dict={
                "address": Address,
                "area": Area,
            }
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Review(Item):

    name: Optional[str] = None
    reviewBody: Optional[str] = None
    reviewRating: Optional[Rating] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    votedHelpful: Optional[int] = None
    votedUnhelpful: Optional[int] = None
    isVerified: Optional[bool] = None
    probability: Optional[float] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(item, from_dict={"reviewRating": Rating})
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class Reviews(Item):

    url: Optional[str] = None
    reviews: List[Review] = attr.Factory(list)
    paginationNext: Optional[PaginationLink] = None
    paginationPrevious: Optional[PaginationLink] = None

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_list={
                'reviews': Review,
            },
            from_dict={
                'paginationNext': PaginationLink,
                'paginationPrevious': PaginationLink,
            }
        )
        return super().from_dict(new_item)


@attr.s(auto_attribs=True, slots=True)
class MileageFromOdometer(Item):

    value: Optional[int] = None
    unitCode: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class VehicleEngine(Item):

    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class AvailableAtOrFrom(Item):

    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class FuelEfficiency(Item):

    raw: Optional[str] = None


@attr.s(auto_attribs=True, slots=True)
class Vehicle(Item):

    url: Optional[str] = None
    canonicalUrl: Optional[str] = None
    probability: Optional[float] = None
    name: Optional[str] = None
    offers: List[Offer] = attr.Factory(list)
    sku: Optional[str] = None
    mpn: Optional[str] = None
    brand: Optional[str] = None
    breadcrumbs: List[Breadcrumb] = attr.Factory(list)
    mainImage: Optional[str] = None
    images: List[str] = attr.Factory(list)
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    additionalProperty: List[AdditionalProperty] = attr.Factory(list)
    aggregateRating: Optional[Rating] = None
    vehicleIdentificationNumber: Optional[str] = None
    mileageFromOdometer: Optional[MileageFromOdometer] = None
    vehicleTransmission: Optional[str] = None
    fuelType: Optional[str] = None
    vehicleEngine: Optional[VehicleEngine] = None
    color: Optional[str] = None
    vehicleInteriorColor: Optional[str] = None
    availableAtOrFrom: Optional[AvailableAtOrFrom] = None
    numberOfDoors: Optional[int] = None
    vehicleSeatingCapacity: Optional[int] = None
    fuelEfficiency: List[FuelEfficiency] = attr.Factory(list)

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        new_item = _apply_types(
            item,
            from_dict={
                'aggregateRating': Rating,
                'mileageFromOdometer': MileageFromOdometer,
                'vehicleEngine': VehicleEngine,
                'availableAtOrFrom': AvailableAtOrFrom,
            },
            from_list={
                'additionalProperty': AdditionalProperty,
                'breadcrumbs': Breadcrumb,
                'offers': Offer,
                'fuelEfficiency': FuelEfficiency,
            }
        )
        return super().from_dict(new_item)



from types import MappingProxyType
from typing import Any, KeysView, Iterator

from itemadapter.adapter import AttrsAdapter

from autoextract_poet.items import Item


class AutoExtractAdapter(AttrsAdapter):
    """
    ``ItemAdapter`` for AutoExtract poet items that deals transparently with
    both the ``attr.s`` defined fields and the rest of unknown
    fields received at item initialization, offering an unified view over
    all the item fields.

    The utility is twofold. Firstly, it serves to ensure the
    pass-through of new fields from the API even if ``autoextract-poet``
    item definitions have not been yet updated. In other words,
    it can be used to create spiders that preserve all the data coming
    from the API even if they don't have updated item definitions.

    Secondly, it offers a common interface to access and modify both kind of
    fields (known and unknown).

    Remember that this adapter should be enabled by invoking::

        ItemAdapter.ADAPTER_CLASSES.appendleft(AutoExtractAdapter)
    """
    def __init__(self, item: Any) -> None:
        super().__init__(item)

    @classmethod
    def is_item(cls, item: Any) -> bool:
        return isinstance(item, Item)

    def get_field_meta(self, field_name: str) -> MappingProxyType:
        if field_name in self._fields_dict:
            return self._fields_dict[field_name].metadata  # type: ignore
        else:
            return MappingProxyType({})

    def field_names(self) -> KeysView:
        return KeysView({**self._fields_dict, **self.item._unknown_fields_dict})

    def __getitem__(self, field_name: str) -> Any:
        if field_name in self._fields_dict:
            return getattr(self.item, field_name)
        elif field_name in self.item._unknown_fields_dict:
            return self.item._unknown_fields_dict[field_name]
        raise KeyError(field_name)

    def __setitem__(self, field_name: str, value: Any) -> None:
        if field_name in self._fields_dict:
            setattr(self.item, field_name, value)
        else:
            self.item._unknown_fields_dict[field_name] = value

    def __delitem__(self, field_name: str) -> None:
        if field_name in self._fields_dict:
            try:
                delattr(self.item, field_name)
            except AttributeError:
                raise KeyError(field_name)
        elif field_name in self.item._unknown_fields_dict:
            del self.item._unknown_fields_dict[field_name]
        else:
            raise KeyError(f"Object of type {self.item.__class__.__name__} does " +
                           f"not contain a field with name {field_name}")

    def __iter__(self) -> Iterator:
        fields = [attr for attr in self._fields_dict if hasattr(self.item, attr)]
        fields.extend(attr for attr in self.item._unknown_fields_dict if attr not in fields)
        return iter(fields)



import sys
from typing import Type, Optional, Tuple, Dict, Callable, Any
from weakref import WeakKeyDictionary

import attr


# Caches the attribute names for attr.s classes
CLASS_ATTRS: WeakKeyDictionary = WeakKeyDictionary()


def split_in_unknown_and_known_fields(data: Optional[dict], item_cls: Type) -> Tuple[Dict, Dict]:
    """
    Return a pair of dicts. The first one contains those elements not belonging to the
    attr class ``item_cls``. The second one contains the rest. That is, those
    attributes not belonging to ``item_cls`` class
    """
    data = data or {}
    if not attr.has(item_cls):
        raise ValueError(f"The cls {item_cls} is not attr.s class")
    if not item_cls in CLASS_ATTRS:
        CLASS_ATTRS[item_cls] = {field.name for field in attr.fields(item_cls)}
    unknown, known = split_dict(data, lambda k: k in CLASS_ATTRS[item_cls])
    return unknown, known


def split_dict(dict: Dict, key_pred: Callable[[Any], Any]) -> Tuple[Dict, Dict]:
    """
    Splits the dictionary in two. The first dict contains the records
    for which the key predicate is False and the second dict contains the rest

    >>> split_dict({}, lambda k: False)
    ({}, {})
    >>> split_dict(dict(a=1, b=2, c=3), lambda k: k != 'a')
    ({'a': 1}, {'b': 2, 'c': 3})
    """
    yes, no = {}, {}
    for k, v in dict.items():
        if key_pred(k):
            yes[k] = v
        else:
            no[k] = v
    return (no, yes)


def export(fn):
    """
    Decorator that includes the decorated element into the
    ``__all__`` variable in the module. Useful to control
    what is imported when ``import * from <module>`` is used.
    """
    mod = sys.modules[fn.__module__]
    if hasattr(mod, '__all__'):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn


from typing import ClassVar, Generic, Optional, TypeVar, Type

import attr

from autoextract_poet.items import (
    Article,
    Item,
    Product, ProductList, ArticleList, Comments, ForumPosts, JobPosting,
    RealEstate, Reviews, Vehicle,
)
from autoextract_poet.util import export


@export
@attr.s(auto_attribs=True)
class AutoExtractHtml:
    """A container for URL and HTML content retrieved from AutoExtract.

    ``url`` should be an URL of the response (after all redirects),
    not an URL of the request, if possible.

    ``html`` should be browser HTML in unicode
    """
    url: str
    html: str


T = TypeVar("T", bound=Item)


@attr.s(auto_attribs=True)
class AutoExtractData(Generic[T]):
    """Container for AutoExtract data.

    Should not be used directly by providers.
    Use derived classes like AutoExtractArticleData and similar.

    API responses are wrapped in a JSON array
    (this is to facilitate query batching)
    but we're receiving single responses here..

    https://docs.zyte.com/automatic-extraction.html#responses
    """
    page_type: ClassVar[str]
    data: dict

    @property
    def item_class(self):
        return get_item_class(self)

    def to_item(self) -> Optional[T]:
        return self.item_class.from_dict(self.data[self.page_type])


def get_item_class(page_input_cls: Type[AutoExtractData]) -> Type[Item]:
    """ Return item class for the page input class.

    >>> get_item_class(AutoExtractArticleData) is Article
    True
    >>> get_item_class(AutoExtractProductData) is Product
    True
    >>> get_item_class(AutoExtractData) is T
    True
    """
    return page_input_cls.__orig_bases__[0].__args__[0]  # type: ignore[attr-defined]


@export
@attr.s(auto_attribs=True)
class AutoExtractArticleData(AutoExtractData[Article]):
    """Container for AutoExtract Article data.

    https://docs.zyte.com/automatic-extraction/article.html
    """
    page_type = "article"


@export
@attr.s(auto_attribs=True)
class AutoExtractProductData(AutoExtractData[Product]):
    """Container for AutoExtract Product data.

    https://docs.zyte.com/automatic-extraction/product.html
    """
    page_type = "product"


@export
@attr.s(auto_attribs=True)
class AutoExtractProductListData(AutoExtractData[ProductList]):
    """Container for AutoExtract Product list data.

    https://docs.zyte.com/automatic-extraction/product-list.html
    """
    page_type = "productList"


@export
@attr.s(auto_attribs=True)
class AutoExtractArticleListData(AutoExtractData[ArticleList]):
    """Container for AutoExtract Article list data.

    https://docs.zyte.com/automatic-extraction/article-list.html
    """
    page_type = "articleList"


@export
@attr.s(auto_attribs=True)
class AutoExtractCommentsData(AutoExtractData[Comments]):
    """Container for AutoExtract Comments data.

    https://docs.zyte.com/automatic-extraction/comment.html
    """
    page_type = "comments"


@export
@attr.s(auto_attribs=True)
class AutoExtractForumPostsData(AutoExtractData[ForumPosts]):
    """Container for AutoExtract Forum Posts data.

    https://docs.zyte.com/automatic-extraction/forum-post.html
    """
    page_type = "forumPosts"


@export
@attr.s(auto_attribs=True)
class AutoExtractJobPostingData(AutoExtractData[JobPosting]):
    """Container for AutoExtract Job Posting data.

    https://docs.zyte.com/automatic-extraction/job-posting.html
    """
    page_type = "jobPosting"


@export
@attr.s(auto_attribs=True)
class AutoExtractRealEstateData(AutoExtractData[RealEstate]):
    """Container for AutoExtract Real Estate data.

    https://docs.zyte.com/automatic-extraction/real-estate.html
    """
    page_type = "realEstate"


@export
@attr.s(auto_attribs=True)
class AutoExtractReviewsData(AutoExtractData[Reviews]):
    """Container for AutoExtract Reviews data.

    https://docs.zyte.com/automatic-extraction/review.html
    """
    page_type = "reviews"


@export
@attr.s(auto_attribs=True)
class AutoExtractVehicleData(AutoExtractData[Vehicle]):
    """Container for AutoExtract Vehicle data.

    https://docs.zyte.com/automatic-extraction/vehicle.html
    """
    page_type = "vehicle"



from setuptools import setup, find_packages


setup(
    name='autoextract-poet',
    version='0.4.0',
    description='web-poet definitions for AutoExtract API',
    long_description=open('README.rst').read(),
    long_description_content_type="text/x-rst",
    author='Zyte Group Ltd',
    author_email='info@zyte.com',
    url='https://github.com/scrapinghub/autoextract-poet',
    packages=find_packages(exclude=['tests',]),
    install_requires=[
        'attrs',
        'itemadapter',
        'web-poet',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)
